# -*- coding: utf-8 -*-

from __future__ import division
from safrac import jit, np


@jit(nopython=True, cache=True, nogil=True)
def _generate(fractal, center, dims, beta, lowcut, highcut):

    for idx in np.ndindex(dims):

        position = np.array(idx)

        r = np.sqrt(np.dot(position-center, position-center))
        if r < 1.e-5: r = 1.e-5 

        fractal[idx] = fractal[idx]/(r**beta)
        fractal[idx] *= (r >= lowcut)
        fractal[idx] *= (r <= highcut)


class FractalGenerator(object):

    def __init__(self, dimension, lowcut=None, highcut=None, seed=None):

        # dimension: Fractal dimension D_2

        self.beta = -(2. * dimension - 7.) / 2. # http://dx.doi.org/10.1017/CBO9781139174695, Fourier Spectral Approach, Eq. 7.75
        self.R = np.random.RandomState(seed)
        self.lowcut = lowcut
        self.highcut = highcut

    def generate(self, dims):

        center = np.asarray(dims)/2.
        fractal = np.fft.fftn(self.R.randn(*dims))

        lowcut = self.lowcut
        highcut = self.highcut

        if self.lowcut is None:
            lowcut = 0.

        if self.highcut is None:
            highcut = 1.e5

        _generate(fractal, center, dims, self.beta, lowcut, highcut)

        fractal = np.real(np.fft.ifftn(np.fft.ifftshift(fractal)))
        fractalmin = np.amin(fractal)
        fractalmax = np.amax(fractal)
        fractalrange = fractalmax - fractalmin

        return (fractal-fractalmin)/fractalrange




"""
.. module:: safrac 
   :platform: Unix, Windows, Macos
   :synopsis: Self Affine Fractal Volume Filling/Filtering.

.. moduleauthor:: Christoph Statz <christoph.statz@tu-dresden.de>


"""

import numpy as np

try:
    from numba import jit, float64, complex128, int_, void 
except:
    def jit(*jit_args, **jit_kwargs):
        def decorator(function):
            def wrapper(*args, **kwargs):
                return function(*args, **kwargs)
            return wrapper
        return decorator

np.seterr(divide='raise')


from .generator import FractalGenerator
from .mask import FractalMask



from __future__ import division

from safrac import jit, np
from safrac import FractalGenerator


@jit(nopython=True, cache=True, nogil=True)
def _fill_volume(vol, distances):

    for i in range(distances.shape[0]):
        for j in range(distances.shape[1]):
            vol[i, j, 0:int(distances[i, j])] = 1


class FractalMask(object):
    # Not yet done.

    def __init__(self, *args, **kwargs):

        self.fractal_generator = FractalGenerator(*args, **kwargs)

    def generate(self, dims, fill_range, fill_offset=0):

        fractal_dims = dims[0:2]
        distances = self.fractal_generator.generate(fractal_dims)
        distances = np.round(distances * fill_range) + fill_offset

        mask = np.zeros(dims)
        _fill_volume(mask, distances)

        return mask
